class Filter:
    # TODO: Add comparisons
    pass


class Equalizer(Filter):
    def __init__(self, band, gain):
        pass


class Karaoke(Filter):
    def __init__(self, *, level, mono_level, filter_band, filter_width):
        self.level = level
        self.mono_level = mono_level
        self.filter_band = filter_band
        self.filter_width = filter_width

        self._payload = {"karaoke": {"level": self.level,
                                     "monoLevel": self.mono_level,
                                     "filterBand": self.filter_band,
                                     "filterWidth": self.filter_width}}


class Timescale(Filter):
    def __init__(self, *, speed, pitch, rate):
        if not speed > 0:
            pass

        if not pitch > 0:
            pass

        if not rate > 0:
            pass

        self.speed = speed
        self.pitch = pitch
        self.rate = rate

        self._payload = {"timescale": {"speed": self.speed,
                                       "pitch": self.pitch,
                                       "rate": self.rate}}


class Tremolo(Filter):
    def __init__(self, *, frequency, depth):
        if not frequency > 0:
            pass

        if not 0 < depth <= 1:
            pass

        self.frequency = frequency
        self.depth = depth

        self._payload = {"tremolo": {"frequency": self.frequency,
                                     "depth": self.depth}}


class Vibrato(Filter):
    def __init__(self, *, frequency, depth):
        if not 0 < frequency <= 14:
            pass

        if not 0 < depth <= 1:
            pass

        self.frequency = frequency
        self.depth = depth

        self._payload = {"vibrato": {"frequency": self.frequency,
                                     "depth": self.depth}}



class GraniteException(Exception):
    """Base of all exceptions."""


class NodesUnavailable(GraniteException):
    """Either no nodes are currently active or all current nodes are not available."""


class InvalidCredientials(GraniteException):
    """Invalid credentials were passed."""


class HTTPException(GraniteException):
    """HTTP request failed."""



import asyncio
import logging

from discord.ext import commands

from .events import Event
from .exceptions import NodesUnavailable
from .node import Node
from .player import Player

log = logging.getLogger(__name__)


class Client:
    """The Python implementation of a client to connect an Andesite node.

    Attributes
    ----------
    bot: :class:`commands.Bot`
    loop: :class:`asyncio.BaseEventLoop`
    nodes: Dict[:class:`str`, :class:`.Node`"""
    _event_hooks = {}

    def __init__(self, bot: commands.Bot):

        self.bot = bot
        self.loop = bot.loop or asyncio.get_event_loop()

        self.nodes = {}

        bot.add_listener(self.update_handler, "on_socket_response")

    def __repr__(self):
        player_count = len(self.players.values())
        return f"<GraniteClient player_count={player_count}>"

    @property
    def players(self):
        """Returns a mapping of guilds and their respective player.

        Returns
        -------
        Dict[:class:`int`, :class:`.Player`"""
        return self._get_players()

    def _get_players(self):
        players = []

        for node in self.nodes.values():
            players.extend(node.players.values())

        return {player.guild_id: player for player in players}

    async def start_node(
        self, host: str, port: int, *, rest_uri: str, password: str, identifier
    ):
        """Prepares a new music node.

        Parameters
        ----------
        host: :class:`str`
        port: :class:`int`
        rest_uri: :class:`str`
        password: Optional[:class:`str`]
        identifier: :class:`str`

        Returns
        -------
        :class:`.Node`"""
        await self.bot.wait_until_ready()

        node = Node(
            host,
            port,
            self.bot.user.id,
            client=self,
            rest_uri=rest_uri,
            password=password,
            identifier=identifier,
        )
        await node.connect(self.bot)

        node.available = True

        self.nodes[identifier] = node

        return node

    def get_player(self, guild_id: int, cls=None):
        """Gets a player corresponding to the guild.

        Parameters
        ----------
        guild_id: :class:`int`
        cls: Optional[:class:`.Player`]
            An optional subclass of :class:`.Player`."""
        try:
            return self.players[guild_id]
        except KeyError:
            pass

        if not self.nodes:
            raise NodesUnavailable("No nodes available")

        nodes = list(self.nodes.values())
        if not cls:
            cls = Player

        player = cls(self.bot, guild_id, nodes[0])
        nodes[0].players[guild_id] = player

        return player

    async def update_handler(self, data):
        if not data:
            return

        if data["t"] == "VOICE_SERVER_UPDATE":
            guild_id = int(data["d"]["guild_id"])
            try:
                player = self.players[guild_id]
            except KeyError:
                pass
            else:
                await player._voice_server_update(data["d"])

        elif data["t"] == "VOICE_STATE_UPDATE":
            # logger.warning(data)
            if int(data["d"]["user_id"]) != self.bot.user.id:
                return

            guild_id = int(data["d"]["guild_id"])
            try:
                player = self.players[guild_id]
            except KeyError:
                pass

            else:
                await player._voice_state_update(data["d"])
        else:
            return

    async def dispatch(self, event: Event):
        """Dispatches events, WIP.

        Parameters
        ----------
        event: :class:`.Event`
            The event to dispatch.
        """
        event_name = "andesite_" + event.name
        self.bot.dispatch(event_name, event)



"""
Made by twitch#7443
"""
__version__ = "0.0.13a"


from .client import Client
from .events import *
from .exceptions import *
from .filters import *
from .objects import Track, Playlist
from .player import Player
from .websocket import WebSocket



import logging

import aiohttp
from discord.ext import commands

from .objects import Playlist, Track
from .websocket import WebSocket

log = logging.getLogger(__name__)


class Node:
    def __init__(
            self,
            host: str,
            port: int,
            user_id: int,
            *,
            client,
            rest_uri: str,
            password,
            identifier,
            shard_id=None
    ):
        self.rest_uri = rest_uri
        self.host = host
        self.port = port
        self.user_id = user_id
        self.identifier = identifier
        self.password = password
        self.shard_id = shard_id

        self._client = client
        self.session = None

        self.players = {}

        self._websocket = None

        self.available = True

    def __repr__(self):
        p_count = len(self.players.keys())
        return f"<GraniteNode player_count={p_count} available={self.available} shard_id={self.shard_id}>"

    def __str__(self):
        return self.identifier

    async def connect(self, bot: commands.Bot):
        self._websocket = WebSocket(
            bot, self.host, self.port, self.password, self
        )

        self.session = aiohttp.ClientSession(loop=self._client.loop)
        await self._websocket._connect()
        log.debug(f"NODE | Connected {self.__repr__()}")

    async def get_tracks(self, query: str):
        password = "null" if not self.password else self.password

        async with self.session.get(f"{self.rest_uri}/loadtracks", params=dict(identifier=query),
                                    headers={"Authorization": password}) as response:
            data = await response.json()

        if data.get('tracks', None) is None:
            log.info(f"REST | No results found for <{query}>")
            return None

        if data['loadType'] == 'PLAYLIST_LOADED':
            return Playlist(data=data)

        tracks = []
        for track in data["tracks"]:
            tracks.append(Track(_id=track["track"], data=track["info"]))

        log.debug(f"REST | Found <{len(tracks)}> results for query: <{query}>")

        return tracks



class Track:
    """The track gotten from Andesite.

    Attributes
    ----------
    id: :class:`str`
        Not the ID found in YouTube, use :attr:`.yt_id`
    data: :class:`dict`
    title: :class:`str`
    author: :class:`str`
    length: :class:`str`
        The length of the track in ``milliseconds``.
    uri: :class:`str`
    is_stream: :class:`bool`
        Whether the current track is a live stream.
    is_seekable: :class:`bool`
        Whether the current track can be seeked.
    position: :class:`int`
    """
    def __init__(self, _id, data: dict):
        self.id = _id
        self.data = data

        self.title = data.get("title")
        self.author = data.get("author")
        self.length = data.get("length")
        self.yt_id = data.get("identifier")
        self.uri = data.get("uri")
        self.is_stream = data.get("isStream")
        self.is_seekable = data.get("isSeekable")
        self.position = data.get("position")

    def __str__(self):
        return self.title


    def __repr__(self):
        return "<Track length={0.length} is_stream={0.is_stream}>".format(self)


class Playlist:
    """Similar to :class:`.Track`, but a playlist containing more tracks.

    Attributes
    ----------
    data: :class:`dict`
    name: :class:`str`
    tracks: List[:class:`.Track`]
    """
    def __init__(self, data: dict):
        self.data = data

        self.name = data['playlistInfo']['name']

        self.tracks = [Track(_id = track['track'], data = track['info']) for track in data['tracks']]


    def __str__(self):
        return self.name


    def __repr__(self):
        return "<Playlist name={0.name}>".format(self)



import json
import logging

import websockets
from discord.ext import commands

from . import events

log = logging.getLogger(__name__)


class WebSocket:
    def __init__(
        self, bot: commands.Bot, host: str, port: int, password: str, node
    ):
        self.bot = bot
        self.host = host
        self.port = port
        self.password = password
        self.user_id = bot.user.id
        self._connection_id = None

        self.metadata = None

        self._ws = None
        self._task = None
        self._closed = False

        self.headers = {
            "Authorization": password,
            "Andesite-Resume-Id": self._connection_id,
            "User-Id": self.user_id
        }

        self._node = node

    @property
    def is_connected(self):
        return self._ws is not None and self._ws.open

    async def _connect(self):
        await self.bot.wait_until_ready()

        uri = f"ws://{self.host}:{self.port}/websocket"

        try:
            self._ws = await websockets.connect(uri=uri, extra_headers=self.headers)
        except Exception as er:
            return log.warning(er)

        if not self._task:
            self._task = self.bot.loop.create_task(self._listen())

        self._closed = False

        log.debug(f"WEBSOCKET | Connection established to {uri}")

    async def _listen(self):
        while True:
            try:
                data = await self._ws.recv()
            except websockets.ConnectionClosed as e:
                if e.code == 4001:
                    log.warning("WEBSOCKET | Incorrect authentication")
                    break

                elif e.code == 4002:
                    log.warning(
                        f"WEBSOCKET | Closed with code 4002, attempting reconnect"
                    )
                    await self._connect()
                    break

                else:
                    log.warning(f"WEBSOCKET | Closed with code {e.code}")
                    break

            if data:
                data = json.loads(data)

                op = data.get("op", None)
                if op == "connection-id":
                    self._connection_id = data.get("id")
                    log.debug(
                        f"WEBSOCKET | Received connection-id of {self._connection_id}"
                    )

                elif op == "metadata":
                    self.metadata = data["data"]
                    log.debug("WEBSOCKET | Received metadata payload.")

                elif op == "player-update":
                    try:
                        await self._node.players[int(data["guildId"])].update_state(
                            data
                        )
                    except Exception as er:
                        log.warning(f"WEBSOCKET | Error in player-state data {data}")

                elif op == "event":
                    await self._event_dispatcher(data)

                else:
                    log.debug(f"WEBSOCKET | Unknown OP {op} returned {data}")

    async def _event_dispatcher(self, data):
        player = self._node.players[
            int(data["guildId"])
        ]  # getting the player that all events use, apart from websocket close.

        event = getattr(events, data["type"], None)

        if event is None:
            log.debug("WEBSOCKET | Unknown event %s, discarding", data["type"])
            return
        
        log.debug("WEBSOCKET | Dispatching event %s for guildId %s", data['type'], data['guildId'])

        await self._node._client.dispatch(event(player, data))

    async def _send(self, **data):
        if self.is_connected:
            log.debug(f"WEBSOCKET | Sending payload {data}")
            await self._ws.send(json.dumps(data))
